@@ -14,6 +14,7 @@ import (
 	"net/url"
 	"os"
 	"path/filepath"
+	"strings"
 	"testing"
 	"time"
 
@@ -631,6 +632,78 @@ func TestRolesAdjustment_ReplaceVoterHonorFailureDomain(t *testing.T) {
 	assert.Equal(t, client.Voter, cluster[5].Role)
 }
 
+// If cluster is imbalanced (all voters in one failure domain), roles get re-shuffled.
+func TestRolesAdjustment_ImbalancedFailureDomain(t *testing.T) {
+	n := 8
+	apps := make([]*app.App, n)
+	cleanups := make([]func(), n)
+
+	for i := 0; i < n; i++ {
+		addr := fmt.Sprintf("127.0.0.1:900%d", i+1)
+		// Half of the nodes will go to failure domain 0 and half on failure domain 1
+		fd := 0
+		if i > n/2 {
+			fd = 1
+		}
+		options := []app.Option{
+			app.WithAddress(addr),
+			app.WithRolesAdjustmentFrequency(4 * time.Second),
+			app.WithFailureDomain(uint64(fd)),
+		}
+		if i > 0 {
+			options = append(options, app.WithCluster([]string{"127.0.0.1:9001"}))
+		}
+
+		// Nodes on failure domain 0 are started first so all voters are initially there.
+		app, cleanup := newApp(t, options...)
+
+		require.NoError(t, app.Ready(context.Background()))
+
+		apps[i] = app
+		cleanups[i] = cleanup
+	}
+
+	for i := 0; i < n; i++ {
+		defer cleanups[i]()
+	}
+
+	for i := 0; i < n; i++ {
+		cli, err := apps[i].Client(context.Background())
+		require.NoError(t, err)
+		require.NoError(t, cli.Weight(context.Background(), uint64(n-i)))
+		defer cli.Close()
+	}
+
+	time.Sleep(18 * time.Second)
+
+	cli, err := apps[0].Leader(context.Background())
+	require.NoError(t, err)
+	defer cli.Close()
+
+	cluster, err := cli.Cluster(context.Background())
+	require.NoError(t, err)
+
+	domain := map[int]bool{
+		0: false,
+		1: false,
+	}
+	for i := 0; i < n; i++ {
+		// We know we have started half of the nodes in failure domain 0 and the other half on failure domain 1
+		fd := 0
+		if i > n/2 {
+			fd = 1
+		}
+		if cluster[i].Role == client.Voter {
+			domain[fd] = true
+		}
+	}
+
+	// All domain must have a voter
+	for _, voters := range domain {
+		assert.True(t, voters)
+	}
+}
+
 // If a voter goes offline, another node takes its place. Preference will be
 // given to candidates with lower weights.
 func TestRolesAdjustment_ReplaceVoterHonorWeight(t *testing.T) {
@@ -870,12 +943,26 @@ func TestOpen(t *testing.T) {
 	assert.NoError(t, err)
 }
 
+// Open a database with disk-mode on a fresh one-node cluster.
+func TestOpenDisk(t *testing.T) {
+	app, cleanup := newApp(t, app.WithAddress("127.0.0.1:9000"), app.WithDiskMode(true))
+	defer cleanup()
+
+	db, err := app.Open(context.Background(), "test")
+	require.NoError(t, err)
+	defer db.Close()
+
+	_, err = db.ExecContext(context.Background(), "CREATE TABLE foo(n INT)")
+	assert.NoError(t, err)
+}
+
 // Test some setup options
 func TestOptions(t *testing.T) {
 	options := []app.Option{
 		app.WithAddress("127.0.0.1:9000"),
 		app.WithNetworkLatency(20 * time.Millisecond),
 		app.WithSnapshotParams(dqlite.SnapshotParams{Threshold: 1024, Trailing: 1024}),
+		app.WithTracing(client.LogDebug),
 	}
 	app, cleanup := newApp(t, options...)
 	defer cleanup()
@@ -954,7 +1041,7 @@ func newApp(t *testing.T, options ...app.Option) (*app.App, func()) {
 // TestExternalConn creates a 3-member cluster using external http connections
 // and ensures the cluster is successfully created, and that the connection is
 // handled manually.
-func TestExternalConn(t *testing.T) {
+func TestExternalConnWithTCP(t *testing.T) {
 	externalAddr1 := "127.0.0.1:9191"
 	externalAddr2 := "127.0.0.1:9292"
 	externalAddr3 := "127.0.0.1:9393"
@@ -1027,6 +1114,83 @@ func TestExternalConn(t *testing.T) {
 	assert.Equal(t, client.Voter, cluster[2].Role)
 }
 
+// TestExternalPipe creates a 3-member cluster using net.Pipe
+// and ensures the cluster is successfully created, and that the connection is
+// handled manually.
+func TestExternalConnWithPipe(t *testing.T) {
+	externalAddr1 := "first"
+	externalAddr2 := "second"
+	externalAddr3 := "third"
+	acceptCh1 := make(chan net.Conn)
+	acceptCh2 := make(chan net.Conn)
+	acceptCh3 := make(chan net.Conn)
+
+	dialChannels := map[string]chan net.Conn{
+		externalAddr1: acceptCh1,
+		externalAddr2: acceptCh2,
+		externalAddr3: acceptCh3,
+	}
+
+	dialFunc := func(_ context.Context, addr string) (net.Conn, error) {
+		client, server := net.Pipe()
+
+		dialChannels[addr] <- server
+
+		return client, nil
+	}
+
+	app1, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr1), app.WithExternalConn(dialFunc, acceptCh1))
+	defer cleanup()
+
+	app2, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr2), app.WithExternalConn(dialFunc, acceptCh2), app.WithCluster([]string{externalAddr1}))
+	defer cleanup()
+
+	require.NoError(t, app2.Ready(context.Background()))
+
+	app3, cleanup := newAppWithNoTLS(t, app.WithAddress(externalAddr3), app.WithExternalConn(dialFunc, acceptCh3), app.WithCluster([]string{externalAddr1}))
+	defer cleanup()
+
+	require.NoError(t, app3.Ready(context.Background()))
+
+	// Get a client from the first node (likely the leader).
+	cli, err := app1.Leader(context.Background())
+	require.NoError(t, err)
+	defer cli.Close()
+
+	// Ensure entries exist for each cluster member.
+	cluster, err := cli.Cluster(context.Background())
+	require.NoError(t, err)
+	assert.Equal(t, externalAddr1, cluster[0].Address)
+	assert.Equal(t, externalAddr2, cluster[1].Address)
+	assert.Equal(t, externalAddr3, cluster[2].Address)
+
+	// Every cluster member should be a voter.
+	assert.Equal(t, client.Voter, cluster[0].Role)
+	assert.Equal(t, client.Voter, cluster[1].Role)
+	assert.Equal(t, client.Voter, cluster[2].Role)
+}
+
+func TestParallelNewApp(t *testing.T) {
+	t.Parallel()
+	for i := 0; i < 100; i++ {
+		i := i
+		t.Run(fmt.Sprintf("run-%d", i), func(tt *testing.T) {
+			tt.Parallel()
+			// TODO: switch this to tt.TempDir when we switch to
+			tmpDir := filepath.Join(os.TempDir(), strings.ReplaceAll(tt.Name(), "/", "-"))
+			require.NoError(tt, os.MkdirAll(tmpDir, 0700))
+			dqApp, err := app.New(tmpDir,
+				app.WithAddress(fmt.Sprintf("127.0.0.1:%d", 10200+i)),
+			)
+			require.NoError(tt, err)
+			defer func() {
+				_ = dqApp.Close()
+				_ = os.RemoveAll(tmpDir)
+			}()
+		})
+	}
+}
+
 func newAppWithDir(t *testing.T, dir string, options ...app.Option) (*app.App, func()) {
 	t.Helper()
 
@@ -1112,3 +1276,79 @@ func newDir(t *testing.T) (string, func()) {
 
 	return dir, cleanup
 }
+
+func Test_TxRowsAffected(t *testing.T) {
+	app, cleanup := newAppWithNoTLS(t, app.WithAddress("127.0.0.1:9001"))
+	defer cleanup()
+
+	err := app.Ready(context.Background())
+	require.NoError(t, err)
+
+	db, err := app.Open(context.Background(), "test")
+	require.NoError(t, err)
+	defer db.Close()
+
+	_, err = db.ExecContext(context.Background(), `
+CREATE TABLE test (
+	id            TEXT PRIMARY KEY,
+	value         INT
+);`);
+	require.NoError(t, err);
+
+	// Insert watermark
+	err = tx(context.Background(), db, func(ctx context.Context, tx *sql.Tx) error {
+		query := `
+INSERT INTO test
+	(id, value)
+VALUES
+	('id0', -1);
+	`
+		result, err := tx.ExecContext(ctx, query)
+		if err != nil {
+			return err
+		}
+		_, err = result.RowsAffected()
+		if err != nil {
+			return err
+		}
+		return nil
+	})
+	require.NoError(t, err)
+
+	// Update watermark
+	err = tx(context.Background(), db, func(ctx context.Context, tx *sql.Tx) error {
+		query := `
+UPDATE test
+SET
+	value = 1
+WHERE id = 'id0';
+	`
+		result, err := tx.ExecContext(ctx, query)
+		if err != nil {
+			return err
+		}
+		affected, err := result.RowsAffected()
+		if err != nil {
+			return err
+		}
+		if affected != 1 {
+			return fmt.Errorf("expected 1 row affected, got %d", affected)
+		}
+		return nil
+	})
+	require.NoError(t, err)
+}
+
+func tx(ctx context.Context, db *sql.DB, fn func(context.Context, *sql.Tx) error) error {
+	tx, err := db.BeginTx(ctx, nil)
+	if err != nil {
+		return err
+	}
+
+	if err := fn(ctx, tx); err != nil {
+		_ = tx.Rollback()
+		return err
+	}
+
+	return tx.Commit()
+}
