@@ -10,6 +10,7 @@ import (
 
 	"github.com/canonical/go-dqlite"
 	"github.com/canonical/go-dqlite/client"
+	"github.com/canonical/go-dqlite/internal/protocol"
 )
 
 // Option can be used to tweak app parameters.
@@ -139,6 +140,15 @@ func WithRolesAdjustmentFrequency(frequency time.Duration) Option {
 	}
 }
 
+// WithRolesAdjustmentHook will be run each time the roles are adjusted, as
+// controlled by WithRolesAdjustmentFrequency. Provides the current raft leader information
+// as well as the most up to date list of cluster members and their roles.
+func WithRolesAdjustmentHook(hook func(leader client.NodeInfo, cluster []client.NodeInfo) error) Option {
+	return func(o *options) {
+		o.OnRolesAdjustment = hook
+	}
+}
+
 // WithLogFunc sets a custom log function.
 func WithLogFunc(log client.LogFunc) Option {
 	return func(options *options) {
@@ -171,6 +181,17 @@ func WithNetworkLatency(latency time.Duration) Option {
 	}
 }
 
+// WithConcurrentLeaderConns is the maximum number of concurrent connections
+// to other cluster members that will be attempted while searching for the dqlite leader.
+// It takes a pointer to an integer so that the value can be dynamically modified based on cluster health.
+//
+// The default is 10 connections to other cluster members.
+func WithConcurrentLeaderConns(maxConns *int64) Option {
+	return func(o *options) {
+		o.ConcurrentLeaderConns = maxConns
+	}
+}
+
 // WithSnapshotParams sets the raft snapshot parameters.
 func WithSnapshotParams(params dqlite.SnapshotParams) Option {
 	return func(options *options) {
@@ -223,8 +244,10 @@ type options struct {
 	Voters                   int
 	StandBys                 int
 	RolesAdjustmentFrequency time.Duration
+	OnRolesAdjustment        func(client.NodeInfo, []client.NodeInfo) error
 	FailureDomain            uint64
 	NetworkLatency           time.Duration
+	ConcurrentLeaderConns    *int64
 	UnixSocket               string
 	SnapshotParams           dqlite.SnapshotParams
 	DiskMode                 bool
@@ -233,14 +256,17 @@ type options struct {
 
 // Create a options object with sane defaults.
 func defaultOptions() *options {
+	maxConns := protocol.MaxConcurrentLeaderConns
 	return &options{
 		Log:                      defaultLogFunc,
 		Tracing:                  client.LogNone,
 		Voters:                   3,
 		StandBys:                 3,
 		RolesAdjustmentFrequency: 30 * time.Second,
+		OnRolesAdjustment:        func(client.NodeInfo, []client.NodeInfo) error { return nil },
 		DiskMode:                 false, // Be explicit about not enabling disk-mode by default.
 		AutoRecovery:             true,
+		ConcurrentLeaderConns:    &maxConns,
 	}
 }
 
