@@ -7,10 +7,11 @@ package bindings
 #include <assert.h>
 #include <stdint.h>
 #include <signal.h>
+#include <stdbool.h>
 
 #include <dqlite.h>
-#include <raft.h>
-#include <sqlite3.h>
+
+#define RAFT_NOCONNECTION 16
 
 #define EMIT_BUF_LEN 1024
 
@@ -61,14 +62,14 @@ static void setInfo(dqlite_node_info_ext *infos, unsigned i, dqlite_node_id id,
 	info->dqlite_role = role;
 }
 
-static int sqlite3ConfigSingleThread()
-{
-	return sqlite3_config(SQLITE_CONFIG_SINGLETHREAD);
-}
+__attribute__((weak))
+int dqlite_node_set_auto_recovery(dqlite_node *t, bool on);
 
-static int sqlite3ConfigMultiThread()
-{
-	return sqlite3_config(SQLITE_CONFIG_MULTITHREAD);
+static int setAutoRecovery(dqlite_node *t, bool on) {
+	if (dqlite_node_set_auto_recovery == NULL) {
+		return DQLITE_ERROR;
+	}
+	return dqlite_node_set_auto_recovery(t, on);
 }
 
 */
@@ -102,22 +103,16 @@ func init() {
 	C.signal(C.SIGPIPE, C.SIG_IGN)
 }
 
-func ConfigSingleThread() error {
-	if rc := C.sqlite3ConfigSingleThread(); rc != 0 {
-		return protocol.Error{Message: C.GoString(C.sqlite3_errstr(rc)), Code: int(rc)}
-	}
-	return nil
-}
-
-func ConfigMultiThread() error {
-	if rc := C.sqlite3ConfigMultiThread(); rc != 0 {
-		return protocol.Error{Message: C.GoString(C.sqlite3_errstr(rc)), Code: int(rc)}
-	}
-	return nil
-}
-
 // NewNode creates a new Node instance.
 func NewNode(ctx context.Context, id uint64, address string, dir string) (*Node, error) {
+	requiredVersion := dqliteMajorVersion*100 + dqliteMinorVersion
+	// Remove the patch version, as patch versions should be compatible.
+	runtimeVersion := int(C.dqlite_version_number()) / 100
+	if requiredVersion > runtimeVersion {
+		return nil, fmt.Errorf("version mismatch: required version(%d.%d.x) current version(%d.%d.x)",
+			dqliteMajorVersion, dqliteMinorVersion, runtimeVersion/100, runtimeVersion%100)
+	}
+
 	var server *C.dqlite_node
 	cid := C.dqlite_node_id(id)
 
@@ -190,6 +185,22 @@ func (s *Node) SetFailureDomain(code uint64) error {
 	return nil
 }
 
+func (s *Node) EnableDiskMode() error {
+	server := (*C.dqlite_node)(unsafe.Pointer(s.node))
+	if rc := C.dqlite_node_enable_disk_mode(server); rc != 0 {
+		return fmt.Errorf("failed to set disk mode")
+	}
+	return nil
+}
+
+func (s *Node) SetAutoRecovery(on bool) error {
+	server := (*C.dqlite_node)(unsafe.Pointer(s.node))
+	if rc := C.setAutoRecovery(server, C.bool(on)); rc != 0 {
+		return fmt.Errorf("failed to set auto-recovery behavior")
+	}
+	return nil
+}
+
 func (s *Node) GetBindAddress() string {
 	server := (*C.dqlite_node)(unsafe.Pointer(s.node))
 	return C.GoString(C.dqlite_node_get_bind_address(server))
