@@ -15,6 +15,8 @@ import (
 // schema.sh to generate encoding logic for statement parameters.
 type NamedValues = []driver.NamedValue
 
+type NamedValues32 = []driver.NamedValue
+
 // Nodes is a type alias of a slice of NodeInfo. It's used by schema.sh to
 // generate decoding logic for the heartbeat response.
 type Nodes []NodeInfo
@@ -23,7 +25,7 @@ type Nodes []NodeInfo
 type Message struct {
 	words  uint32
 	mtype  uint8
-	flags  uint8
+	schema uint8
 	extra  uint16
 	header []byte // Statically allocated header buffer
 	body   buffer // Message body data.
@@ -45,7 +47,7 @@ func (m *Message) Init(initialBufferSize int) {
 func (m *Message) reset() {
 	m.words = 0
 	m.mtype = 0
-	m.flags = 0
+	m.schema = 0
 	m.extra = 0
 	for i := 0; i < messageHeaderSize; i++ {
 		m.header[i] = 0
@@ -157,21 +159,7 @@ func (m *Message) putFloat64(v float64) {
 	binary.LittleEndian.PutUint64(b.Bytes[b.Offset:], math.Float64bits(v))
 }
 
-// Encode the given driver values as binding parameters.
-func (m *Message) putNamedValues(values NamedValues) {
-	l := len(values)
-	if l > 255 {
-		// safeguard, should have been checked beforehand.
-		panic("too many parameters")
-	}
-
-	n := uint8(l) // N of params
-	if n == 0 {
-		return
-	}
-
-	m.putUint8(n)
-
+func (m *Message) putNamedValuesInner(values NamedValues) {
 	for i := range values {
 		if values[i].Ordinal != i+1 {
 			panic("unexpected ordinal")
@@ -229,12 +217,42 @@ func (m *Message) putNamedValues(values NamedValues) {
 			panic("unsupported value type")
 		}
 	}
+}
 
+// Encode the given driver values as binding parameters.
+func (m *Message) putNamedValues(values NamedValues) {
+	l := len(values)
+	if l == 0 {
+		return
+	} else if l > math.MaxUint8 {
+		// safeguard, should have been checked beforehand.
+		panic("too many parameters")
+	}
+	n := uint8(l)
+
+	m.putUint8(n)
+	m.putNamedValuesInner(values)
+}
+
+// Encode the given driver values as binding parameters, with a 32-bit
+// parameter count (new format).
+func (m *Message) putNamedValues32(values NamedValues) {
+	l := len(values)
+	if l == 0 {
+		return
+	} else if int64(l) > math.MaxUint32 {
+		// safeguard, should have been checked beforehand.
+		panic("too many parameters")
+	}
+	n := uint32(l)
+
+	m.putUint32(n)
+	m.putNamedValuesInner(values)
 }
 
 // Finalize the message by setting the message type and the number
 // of words in the body (calculated from the body size).
-func (m *Message) putHeader(mtype uint8) {
+func (m *Message) putHeader(mtype, schema uint8) {
 	if m.body.Offset <= 0 {
 		panic("static offset is not positive")
 	}
@@ -244,7 +262,7 @@ func (m *Message) putHeader(mtype uint8) {
 	}
 
 	m.mtype = mtype
-	m.flags = 0
+	m.schema = schema
 	m.extra = 0
 
 	m.words = uint32(m.body.Offset) / messageWordSize
@@ -259,7 +277,7 @@ func (m *Message) finalize() {
 
 	binary.LittleEndian.PutUint32(m.header[0:], m.words)
 	m.header[4] = m.mtype
-	m.header[5] = m.flags
+	m.header[5] = m.schema
 	binary.LittleEndian.PutUint16(m.header[6:], m.extra)
 }
 
@@ -274,9 +292,9 @@ func (m *Message) bufferForPut(size int) *buffer {
 	return &m.body
 }
 
-// Return the message type and its flags.
+// Return the message type and its schema version.
 func (m *Message) getHeader() (uint8, uint8) {
-	return m.mtype, m.flags
+	return m.mtype, m.schema
 }
 
 // Read a string from the message body.
@@ -304,18 +322,9 @@ func (m *Message) getString() string {
 func (m *Message) getBlob() []byte {
 	size := m.getUint64()
 	data := make([]byte, size)
-	for i := range data {
-		data[i] = m.getUint8()
-	}
-	pad := 0
-	if (size % messageWordSize) != 0 {
-		// Account for padding
-		pad = int(messageWordSize - (size % messageWordSize))
-	}
-	// Consume padding
-	for i := 0; i < pad; i++ {
-		m.getUint8()
-	}
+	b := m.bufferForGet()
+	defer b.Advance(int(alignUp(size, messageWordSize)))
+	copy(data, b.Bytes[b.Offset:])
 	return data
 }
 
@@ -460,6 +469,12 @@ func (r *Rows) columnTypes(save bool) ([]uint8, error) {
 		r.types = make([]uint8, len(r.Columns))
 	}
 
+	// If there are zero columns, no rows can be encoded or decoded,
+	// so we signal EOF immediately.
+	if len(r.types) == 0 {
+		return r.types, io.EOF
+	}
+
 	// Each column needs a 4 byte slot to store the column type. The row
 	// header must be padded to reach word boundary.
 	headerBits := len(r.types) * 4
@@ -660,3 +675,8 @@ func (r *Rows) ColumnTypes() ([]string, error) {
 
 	return kinds, err
 }
+
+// alignUp rounds n up to a multiple of a. a must be a power of 2.
+func alignUp(n, a uint64) uint64 {
+	return (n + a - 1) &^ (a - 1)
+}
